""" Very simple example of using ETI API (using module :mod:`.eti_api`).

To run this script, you need to create also 'test_credentials.txt' which will contain 3 lines:
    host    (should be: "eti.eset.com")
    username
    password
    OPTIONAL: token
"""

import logging
import os
import sys  # Just to check compatibility with Python 2.7.

import eti_api

TMP_DIR = "TMP"
LOG_PATH = os.path.join(TMP_DIR, "debug.log")

# Some hashes
HASHES = [
    "A7BBC4B4F781E04214ECEBE69A766C76681AA7EB", # notepad.exe
    "D093160B9069CAAADC9E666FB848731B90187DCF"  # malicious qadars sample
]

def _initialize_debug_logging():
    """ Initializes debug logging into LOG_PATH.
    """
    debug_handler = logging.FileHandler(LOG_PATH, "w", encoding="utf-8")
    fmt = "%(levelname)s %(asctime)s %(message)s"
    datefmt = "%Y-%m-%d %H:%M:%S"
    if sys.version_info.major == 2:  # Compatibility with Python 2.7.
        formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)
    else:
        formatter = logging.Formatter(fmt=fmt, datefmt=datefmt, style='%')
    debug_handler.setFormatter(formatter)
    debug_handler.setLevel('DEBUG')
    eti_api.LOGGER.addHandler(debug_handler)
    eti_api.LOGGER.setLevel("DEBUG")

def main():
    """ Executes a few simple use-cases.
    """
    # Create temporary dir and initialize logging.
    if sys.version_info.major == 2:  # Compatibility with Python 2.7.
        if not os.path.isdir(TMP_DIR):
            os.makedirs(TMP_DIR)
    else:
        os.makedirs(TMP_DIR, exist_ok=True)
    _initialize_debug_logging()

    # Load credentials.
    with open("test_credentials.txt") as file:
        host = file.readline().strip()
        username = file.readline().strip()
        password = file.readline().strip()
        token = file.readline()
        if token:
            token = token.strip()
        else:
            token = None

    with eti_api.Connection(username=username, password=password, host=host, token=token) as connection:
        ################################################################################################################
        # Example: submit hashes and download generated Sample Reports.
        ################################################################################################################
        # Generate all reports from given HASHES.
        reports = []
        for hash_ in HASHES:
            report = connection.submit_hash(hash_)
            reports.append(report)
        # Wait until all reports are generated and downloaded them.
        timeout = 600
        for report in reports:
            report = connection.wait_on_report(report, timeout=timeout)
            timeout = int(timeout / 6)  # We can wait less for next reports...
            _download_all_report_stuff(connection, report, os.path.join(TMP_DIR, report["sha"]))

        ################################################################################################################
        # Example: download all phishing and botnet reports from last week.
        ################################################################################################################
        for type_ in ["phish", "botnet"]:
            i = 0
            for report in connection.list_reports(type_=type_, pid="1w"):
                if report["status"] != "finished":
                    continue  # Skip not generated reports.
                i += 1
                _download_all_report_stuff(connection, report, os.path.join(TMP_DIR, "{}_{:02d}".format(type_, i)))

        ################################################################################################################
        # Example: get some yara match, mark it as read and generete and download report from it.
        ################################################################################################################
        # Get all unread matches from last 48 hours.
        ymatches = connection.list_yara_matches(tid="48h", rst=1)
        # Process first match.
        ymatch = ymatches[0]
        # Download the match detail
        connection.get_yara_match_detail(ymatch, file_path=os.path.join(TMP_DIR, "ydetail.xml"))
        # Generate report.
        report = connection.generate_yara_match_report(ymatch)
        report = connection.wait_on_report(report)
        # Report generated => Mark Yara match as done.
        connection.mark_yara_match(ymatch)
        # Download all stuff from generated report.
        _download_all_report_stuff(connection, report, os.path.join(TMP_DIR, "yreport"))

def _download_all_report_stuff(connection, report, base_path):
    """ Download xml, pdf and adds (if available) from given *report* into paths starting with *base_path*.
    """
    for fmt in ["pdf", "xml", "adds"]:
        ext = fmt if fmt != "adds" else "zip"
        connection.get_report(report, fmt, file_path="{}.{}".format(base_path, ext))

if __name__ == "__main__":
    main()
